# 
# Copyright pyPr v1.0b: 2024, John Badger.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

import math
import sys
import os
import os.path
import statistics
import random
import time

#############
# FUNCTIONS #
#############

# Find q for minimum i using small bins
def test_data_i_min(data_q_aList,data_i_aList,nbin):

    num = len(data_q_aList)
    i_min = max(data_i_aList)
    q_for_i_min = max(data_q_aList)
    nbin_half = int(nbin/2)

    i = nbin_half
    while i < num - nbin_half - 1:

        i_mean = statistics.mean(data_i_aList[i-nbin_half:i+nbin_half])

        if i_mean < i_min:
            i_min = i_mean
            q_for_i_min = data_q_aList[i] 

        i = i + 1

    return(q_for_i_min);

# Parse input data from a GNOM file
def read_data_gnom(aFile_in,qmin,qmax,\
              data_q_aList,data_i_aList,data_sd_aList):

    allLines = []

    file = open(aFile_in,'r')
    allLines = file.readlines()
    file.close()

    for eachLine in allLines:

        aList = eachLine.split()
        num_params = len(aList)

        if num_params == 5:

            if ('.' in aList[0]) and ('.' in aList[1]) and ('.' in aList[2]) and \
               ('E' in aList[0]) and ('E' in aList[1]) and ('E' in aList[2]):

                q = float(aList[0])
                i = float(aList[1])            
                sd = float(aList[2])

                if q > qmin and q < qmax:
                    data_q_aList.append(q)
                    data_i_aList.append(i)
                    data_sd_aList.append(sd)
                    
    return;

# Get Rg from the P(r)
def get_rg(pr_aList,r_aList,dmax):

    
    pi = 3.14159
    sum_r2pr = 0.0
    sum_pr = 0.0

    num = len(pr_aList)

    i = 0
    while i < num:

        r = float(r_aList[i])
        pr = float(pr_aList[i])

        if r < dmax:
            sum_r2pr = sum_r2pr + r*r*pr
            sum_pr = sum_pr + pr

        i = i + 1

    rg_pr = math.sqrt(sum_r2pr/(2.0*sum_pr))

    return(rg_pr)

# Check suitable data in the Guinier region.
def check_guinier(data_q_aList,data_i_aList,qmax_rg,qmin_rg):

    num_rg = 0
    num = len(data_q_aList)

    i = 0
    while i < num:
        if data_q_aList[i] > qmin_rg and data_q_aList[i] < qmax_rg:
            num_rg = num_rg + 1

        i = i + 1

    return(num_rg)

# Make smooth truncation of P(r)at dmax using last 10% of P(r)
def smooth_dmax(pr_aList,dmax_pt):

    dmax_start_pt = int(dmax_pt/1.1)
    
    pr_start = pr_aList[dmax_start_pt]
    decrease_step = pr_start/(dmax_pt - dmax_start_pt)

    i = dmax_start_pt + 1
    j = 0
    while i < dmax_pt:
        j = j + 1
        pr_new = pr_start - j*decrease_step
        pr_aList[i] = min(pr_new,pr_aList[i])
        pr_aList[i] = max(0.0,pr_aList[i])

        i = i + 1
        
    pr_aList[dmax_pt]= 0.0

    return;

# Find dmax 
def find_dmax(i_cut,r_aList,pr_aList,rg):

    pr_cut = i_cut*max(pr_aList)
    dmax = 0.0

    # Evaluate dmax based on finding where P(r) approaches zero
    num = len(pr_aList)
    i = 0
    while i < num:
        if r_aList[i] > 2.0*rg:

            if pr_aList[i] < pr_cut:
                dmax = r_aList[i]
                i = num

        i = i + 1

    dmax = 1.05*dmax

    return dmax;

# Create data set filtered by a 4th-order Butterworth filter
def get_filtered_i(data_rect_filter_i_aList,data_rect_i_aList,data_rect_q_aList,qmax_filter):

    num = len(data_rect_q_aList)

    i = 0
    while i < num:

        q = data_rect_q_aList[i]
        i_filtered = data_rect_i_aList[i]/(math.sqrt(1.0+ (q/qmax_filter)**4))
        data_rect_filter_i_aList.append(i_filtered)

        i = i + 1

    return;

# Write summary statistics to a text file
def write_statistics(log_file,aString_time,sample_id,qmin_true,qmax_true,q_for_i_min,\
                     i_ratio,dmax_est,dmax,r_factor,chisq,rg_pr,aFile_in,aFile_in_pr,dif_pr):

    file = open(log_file,'w')

    file.write(aString_time + '\n')
    file.write('ID= ' + str(sample_id) + '\n')
    file.write('Input data file= ' + str(aFile_in) + '\n')
    file.write('Min q from data input= ' + "%6.4f"%(qmin_true) + '\n')
    file.write('Max q from data input= ' + "%6.4f"%(qmax_true) + '\n')
    file.write('Ratio of I(qmax) to I0= ' + "%4.3f"%(i_ratio) + ' %\n')
    file.write('Estimated dmax from P(r)= ' + "%4.1f"%(dmax_est) + '\n')   
    file.write('Used dmax= ' + "%4.1f"%(dmax) + '\n')

    if q_for_i_min > 0.0:
        file.write('Applied Butterworth filter with qthresh= ' + "%6.4f"%(q_for_i_min) + '\n')
    
    file.write('Rg from P(r)= ' + "%5.2f"%(rg_pr) + '\n')
    file.write('R-factor= ' + "%6.4f"%(r_factor) + '\n')
    file.write('ChiSQ= ' + "%5.2f"%(chisq) + '\n')

    if aFile_in_pr != 'none':

        file.write('Comparison P(r) file= ' + str(aFile_in_pr) + '\n')
        file.write('Difference from reference P(r)= ' + "%4.2f"%(100.0*dif_pr) + ' %')

    file.close()

    return;
  
# Convert data units from nm**-1 to A**-1
def convert_q(data_q_aList):

    num = len(data_q_aList)
    i = 0
    while i < num:
        data_q_aList[i] = 0.1*data_q_aList[i]
        i = i + 1

    return;

# Write file of gnom-like output fields file. (Reverse engineered from GNOM examples).
def write_gnomlike(i_and_pr_file,\
                   data_q_aList,data_i_aList,data_sd_aList,data_i_calc_aList,\
                   r_aList,pr_aList,pr_sd_aList,r_limit):

    file = open(i_and_pr_file,'w')

    # The GNOM title line seems to be a key to plotting  P(r)
    file.write('\n')
    file.write('           ####      G N O M-like   Mimics gnom output data fields ####\n')
    

    file.write('\n')
    file.write('\n')

    # Write intensity block header
    file.write('           ####      Experimental Data and Fit                     ####\n')
    file.write('\n')
    file.write('     S          J EXP       ERROR       J REG       I REG\n')
    file.write('\n')

    # Write intensity data block
    num = len(data_i_aList)
    i = 0
    while i < num:

        q = float(data_q_aList[i])
        intensity = float(data_i_aList[i])
        sd = float(data_sd_aList[i])
        ic = float(data_i_calc_aList[i])

        file.write(str("%12.4e"%(q)) + str("%12.4e"%(intensity)) + str("%12.4e"%(sd)) + \
                   str("%12.4e"%(ic)) + str("%12.4e"%(ic)) + '\n')        

        i = i + 1

    # Write P(r) data block header
    file.write('\n')
    file.write('          Distance distribution  function of particle  \n')
    file.write('\n')
    file.write('\n')
    file.write('      R          P(R)      ERROR\n')
    file.write('\n')

    
    # Write P(r) data block
    i = 0
    while i < r_limit:

        r = float(r_aList[i])
        pr = float(pr_aList[i])
        pr_error = float(pr_sd_aList[i])
        
        file.write(str("%12.4e"%(r)) + str("%12.4e"%(pr)) + str("%12.4e"%(pr_error)) + '\n')

        i = i + 1
    
    file.write('')
    file.close()

    return;

# Get fractional differences between two P(r) functions (same step, same length)
def compare_pr(pr_aList,pr_ref_aList):

    num = len(pr_aList)
    dif_pr = 0.0
    sum_pr = 0.0

    i = 0
    while i < num:

        dif_pr = dif_pr + abs(pr_aList[i] - pr_ref_aList[i])
        sum_pr = sum_pr + pr_ref_aList[i]
        
        i = i + 1

    dif = dif_pr/sum_pr

    return(dif);

# Read reference P(r) assuming three-columns of floats
def read_ref_pr(aFile_in_pr,r_ref_aList,pr_ref_aList,pr_sd_ref_aList):

    allLines = []

    file = open(aFile_in_pr,'r')
    allLines = file.readlines()
    file.close()

    for eachLine in allLines:
        
        aList = eachLine.split()
        
        if len(aList) == 3:

            if ('.' in aList[0]) and ('.' in aList[1]) and ('.' in aList[2]):

                r_ref_aList.append(float(aList[0]))
                pr_ref_aList.append(float(aList[1]))
                pr_sd_ref_aList.append(float(aList[2]))

    return;
            
# Discover intensity data file type (CRYSOL, GNOM, general
def find_data_file_type(aFile_in):

    allLines = []

    file = open(aFile_in,'r')
    allLines = file.readlines()
    file.close()

    file_type = 'general'

    for eachLine in allLines:

        aList = eachLine.split()

        if len(aList) > 5 and \
           aList[1] == 'G' and aList[2] == 'N' and aList[3] == 'O' and aList[4] == 'M':
            file_type = 'gnom'

        if len(aList) > 0:
            if aList[0] == 'Dif/Atom/Shape/Bord':
                file_type = 'crysol'

    return(file_type);


# Normalize P(r) so that integral is a preset value
def  normalize_pr(r_aList,pr_aList,pr_sd_aList,set_pr_scale):

    step = r_aList[1]
    rescale = set_pr_scale/(sum(pr_aList)*step)
    
    num = len(pr_aList)
    i = 0
    while i < num:
        pr_aList[i] = rescale*pr_aList[i]
        pr_sd_aList[i] = rescale*pr_sd_aList[i]
        i = i + 1

    return;

# Calculate R-Factor
def get_r_factor(data_i_aList,data_i_test_aList):

    num = len(data_i_aList)
    r_dif = 0.0
    r_sum = 0.0

    i = 0
    while i < num:
        r_dif = r_dif + abs(data_i_aList[i] - data_i_test_aList[i])
        r_sum = r_sum + abs(data_i_aList[i])
        i = i + 1

    return(r_dif/r_sum);

# Extrapolate data using q-4 (marginally helpful)
def q4_extend(data_rect_i_aList,data_rect_q_aList,mean_dq,i_ratio_end):

    num = len(data_rect_i_aList)
    i_last = statistics.mean(data_rect_i_aList[num-6:num-1])
    i_end_extend = i_ratio_end*i_last
    q_last = data_rect_q_aList[num-1]

    scale = i_last*(q_last)**4
    num_add = 0

    q = q_last + mean_dq
    i = 0
    while i < 1000:

        i_extend = scale/q**4

        if i_extend > i_end_extend:

            data_rect_i_aList.append(i_extend)
            data_rect_q_aList.append(q)
            
            q = q + mean_dq
            num_add = num_add + 1

        i = i + 1

    return (num_add,q);

# Flatten P(r) to zero beyond dmax
def pr_bound(pr_aList,r_aList,dmax_pt):

    num = len(pr_aList)

    i = 0
    while i < num:
        if i > dmax_pt:
            pr_aList[i] = 0.0
        i = i + 1

    return;

# Calculate Chi-square
def get_chisq(data_i_aList,data_i_calc_aList,data_sd_aList):

    num_q = len(data_i_aList)
    sum_chisq = 0.0
    chisq = 0.0

    i = 0
    while i < num_q:
        sd = data_sd_aList[i]
        if sd != 0.0:
            dif = data_i_aList[i] - data_i_calc_aList[i]
            sum_chisq = sum_chisq + dif**2/sd**2        
        i = i + 1

    return(chisq);                

# Scale intensities using a multiplier and constant
def scale_i(data1_i_aList,data1_i_calc_aList,data1_q_aList,qmax):

    num_data = len(data1_i_aList)

    S = 0.0
    Sx = 0.0
    Sy = 0.0
    Sxx = 0.0
    Sxy = 0.0

    i = 0
    while i < num_data:

        if data1_q_aList[i] <= qmax:
    
            y = float(data1_i_aList[i])
            x = float(data1_i_calc_aList[i])
            S = S + 1.0
            Sx = Sx + x
            Sy = Sy + y
            Sxx = Sxx + x*x
            Sxy = Sxy + x*y

        i = i + 1

    delta = S*Sxx - Sx*Sx
    a = (Sxx*Sy - Sx*Sxy)/delta
    b = (S*Sxy - Sx*Sy)/delta
        
    return(a,b);

# Interpolated intensity
def get_i_interp(data_q_aList,data_i_aList,q):

    num_q = len(data_i_aList)

    # Find pointers to data values around target q
    i = 0
    while i < num_q:

        if data_q_aList[i] <= q:
            ilow = i
        else:
            ihigh = i
            i = num_q

        i = i + 1

    dq_low = q - data_q_aList[ilow]
    dq_high = data_q_aList[ihigh] - q
    dq_sum = dq_low + dq_high
    
    i_interp = (dq_high*data_i_aList[ilow] + dq_low*data_i_aList[ihigh])/dq_sum

    return(i_interp);
    
# Predicted intensity from Guinier equation
def get_guinier(rg_obs,I0_obs,q):

    guinier = ((q*rg_obs)**2)/3.0
    i_predict = I0_obs*math.exp(-guinier)

    return(i_predict)

# Flat Background intensity adjustment
def shift_background(data_i_aList,min_i):

    num_data = len(data_i_aList)

    i = 0
    while i < num_data:
        data_i_aList[i] = data_i_aList[i] - min_i
        i = i + 1

    return;

# Calculate Rg from appropriate data
def calc_rg(data_q,data_i,qmax_rg,qmin_rg):

    num_q = len(data_q)

    # Standard least-squares fit to Guinier region
    S = 0.0
    Sx = 0.0
    Sy = 0.0
    Sxx = 0.0
    Sxy = 0.0

    i = 0
    while i < num_q:
    
        x = float(data_q[i])

        if (x > qmin_rg) and (x < qmax_rg):
        
            y = float(data_i[i])
            y = math.log(y)
            x = x**2

            S = S + 1.0
            Sx = Sx + x
            Sy = Sy + y
            Sxx = Sxx + x*x
            Sxy = Sxy + x*y

        i = i + 1

    delta = S*Sxx - Sx*Sx
    a = (Sxx*Sy - Sx*Sxy)/delta
    b = (S*Sxy - Sx*Sy)/delta

    rg_internal = math.sqrt(-3.0*b)
    i0_internal = math.exp(a)

    return(rg_internal,i0_internal);

# Apply scale to calculated intensities
def apply_scale(data_i_aList,data_i_calc_aList,a,b):

    num_data = len (data_i_aList)

    i = 0
    while i < num_data:
        data_i_calc_aList[i] = b*data_i_calc_aList[i] + a
        i = i + 1

    return;

# Calculate intensity from P(r) (droping all scaling factors)
def ft_to_intensity(data_q_aList,data_i_calc_aList,r_aList,pr_aList):

    num_q = len(data_q_aList)
    num_r = len(r_aList)
    
    count_q = 0
    while count_q < num_q:

        q = float(data_q_aList[count_q])
        integral = 0.0

        count_r = 1
        while count_r < num_r:

            r = float(r_aList[count_r])
            pr = float(pr_aList[count_r])
            qr = q*r
            integral = integral + pr*math.sin(qr)/qr

            count_r = count_r + 1

        data_i_calc_aList.append(integral)

        count_q = count_q + 1

    return;


# Parse simulated input data from a CRYSOL file
def read_data_crysol(aFile_in,qmin,qmax,\
              data_q_aList,data_i_aList,data_sd_aList):

    allLines = []

    file = open(aFile_in,'r')
    allLines = file.readlines()
    file.close()

    for eachLine in allLines:
        
        aList = eachLine.split()
        num_params = len(aList)

        if num_params == 5 and ('.' in aList[0]):
            q = float(aList[0])
            i = float(aList[1])            
            sd = 0.01*i

            q = float(q)
            if q > qmin and q < qmax:
                data_q_aList.append(q)
                data_i_aList.append(i)
                data_sd_aList.append(sd)

    return;

# Parse input data from general 3-column data file
def read_data(aFile_in,qmin,qmax,\
              data_q_aList,data_i_aList,data_sd_aList):

    allLines = []

    file = open(aFile_in,'r')
    allLines = file.readlines()
    file.close()

    for eachLine in allLines:
        
        aList = eachLine.split()
        num_params = len(aList)

        if num_params == 3 and '.' in aList[0]:
            q = float(aList[0])
            i = float(aList[1])            
            sd = float(aList[2])

            q = float(q)
            if q > qmin and q < qmax:
                data_q_aList.append(q)
                data_i_aList.append(i)
                data_sd_aList.append(sd)

    return;

# Perform Direct Fourier Transform for point P(r)
def calc_pr(data_q_aList,data_reg_i_aList,dq,r):

    pi = 3.141593
    num_q = len(data_q_aList)
        
    factor = r*r/(2.0*pi*pi)
    integral = 0.0
    
    i = 0
    while i < num_q:

        q = data_q_aList[i]
        intensity = data_reg_i_aList[i]
        qr = q*r
        integral = integral + q*q*intensity*dq*math.sin(qr)/qr

        i = i + 1

    pr = factor*dq*integral

    return(pr);

# Write P(r)
def write_pr(pr_file,r_aList,pr_aList,pr_sd_aList,r_limit):

    file = open(pr_file,'w')

    i = 0
    while i < r_limit:

        r = float(r_aList[i])
        pr = float(pr_aList[i])
        pr_error = float(pr_sd_aList[i])
        
        r = ("%12.4e"%(r))
        pr = ("%12.4e"%(pr))
        pr_error = ("%12.4e"%(pr_error))
    
        aString = str(r) + str(pr) + str(pr_error) + '\n'
        file.write(aString)

        i = i + 1

    file.close()        

    return;

# Write intensities
def write_i(i_file,data_q_aList,data_i_aList,data_i_calc_aList):

    num = len(data_i_calc_aList)

    file = open(i_file,'w')

    i = 0
    while i < num:

        q = float(data_q_aList[i])
        intensity = float(data_i_aList[i]) 
        ic = float(data_i_calc_aList[i])

        q = ("%12.4e"%(q))
        intensity = ("%12.4e"%(intensity))
        ic = ("%12.4e"%(ic))
    
        aString = str(q) + str(intensity) + str(ic) + '\n'
        file.write(aString)

        i = i + 1

    file.close()        

    return;

############
# MAIN     #
############

# Opener
version_aString = '\nProgram: pyPr version 1.0b'
print (version_aString)
aString = 'Author: John Badger'
print (aString)
aString = 'Copyright: 2024, John Badger'
print (aString)
aString = 'License: GNU GPLv3'
print (aString)

localtime = time.asctime( time.localtime(time.time()) )
aString_time = 'Starting time: ' + str(localtime) + '\n'
print (aString_time)

# Operating constants and parameters
pi = 3.14159
dmax_in = 0.0
i_ratio_end = 0.1
set_pr_scale = 5000.0
aFile_in = 'none'
aFile_in_pr = 'none'
sample_id = 'not_set'
rg = 0.0
dif_pr = 0.0
rstep = 1.0
qmin = 0.0
qmax = 100.0
qmax_guinier = 0.0
dmax_from_sd = 0.0
num_trial = 20
dq_i_min = 0.05
i_cut = 0.04
nbin = 20

# Get parameter file
aFile_run = input('Enter parameter file name: ')
dirname = os.path.dirname(aFile_run)

# Parse run parameters from a {tag, number} format

if not os.path.isfile(aFile_run):
    aString = 'The input parameter file was not found !'
    print (aString)
    time.sleep(5)
    sys.exit(1)

allLines = []

file = open(aFile_run,'r')
allLines = file.readlines()
file.close()

for eachLine in allLines:
        
    aList = eachLine.split()
        
    if len(aList) == 2:

        if aList[0] == 'data_file':
            aFile_in = os.path.join(dirname,str(aList[1]))

        if aList[0] == 'id':
            sample_id = str(aList[1])       
    
        if aList[0] == 'rg':
            rg = float(aList[1])

        if aList[0] == 'dmax':
            dmax_in = float(aList[1])

        if aList[0] == 'qmin':
            qmin = float(aList[1])

        if aList[0] == 'qmax':
            qmax = float(aList[1])

        if aList[0] == 'ref_pr_file':
            aFile_in_pr = os.path.join(dirname,str(aList[1]))

        if aList[0] == 'pr_step':
            rstep = float(aList[1])

        if aList[0] == 'qmax_guinier':
            qmax_guinier = float(aList[1])

            
# Check required items are present
if aFile_in == 'none':
    aString = 'Input data file name was not given !'
    print (aString)
    time.sleep(5)
    sys.exit(1)

if sample_id == 'not_set':
    aString = 'The identifier for output files was not given !'
    print (aString)
    time.sleep(5)
    sys.exit(1)

if rg == 0.0:
    aString = 'The initial estimate for Rg was not given !'
    print (aString)
    time.sleep(5)
    sys.exit(1)

aString = '\nRun id= ' + str(sample_id)
print (aString)

# Check input parameters are reasonable
if os.path.isfile(aFile_in):
    aString = 'Input data file= ' + str(aFile_in)
    print (aString)
else:
    aString = 'The input data file was not found !'
    print (aString)
    time.sleep(5)
    sys.exit(1)

if aFile_in_pr != 'none':

    if os.path.isfile(aFile_in_pr):
        aString = 'Reference P(r) file= ' + str(aFile_in_pr)
        print (aString)
    else:
        aString = 'The reference P(r) file was not found !'
        print (aString)
        time.sleep(5)
        sys.exit(1)

if rstep < 0.2 or rstep > 2.0:
    aString = 'The value for pr_step is outside normal bounds !'
    print (aString)
    time.sleep(5)
    sys.exit(1)

# Setup output file names
pr_file_start = os.path.join(dirname, (sample_id + '_start_pr.dat'))
pr_file_mod = os.path.join(dirname, (sample_id + '_final_pr.dat'))
pr_file_filter = os.path.join(dirname, (sample_id + '_filter_pr.dat'))
i_file = os.path.join(dirname, (sample_id + '_i.int'))
i_and_pr_file = os.path.join(dirname, (sample_id + '_i_and_pr.out'))
log_file = os.path.join(dirname, (sample_id + '_results.txt'))

# Plots better with PRIMUS using extensiion .int
pr_file_with_ref = os.path.join(dirname, (sample_id + '_with_ref_pr.int'))

max_r = 4.5*rg
r_max = int(max_r/rstep)
qmin_rg = 0.65/rg
qmax_rg = 1.3/rg

if qmax_guinier > 0.001:
    qmax_rg = qmax_guinier

#############################
# Read input intensity data #
#############################

data_q_aList = []
data_i_aList = []
data_sd_aList = []

file_type = find_data_file_type(aFile_in)

if file_type == 'crysol':
    read_data_crysol(aFile_in,qmin,qmax,\
              data_q_aList,data_i_aList,data_sd_aList)
    aString = 'Input data file appears to be from CRYSOL'
    print (aString)
    
if file_type == 'gnom':
    read_data_gnom(aFile_in,qmin,qmax,\
              data_q_aList,data_i_aList,data_sd_aList)
    aString = 'Input data file appears to be from GNOM'
    print (aString)
    
else:    
    read_data(aFile_in,qmin,qmax,\
              data_q_aList,data_i_aList,data_sd_aList)
    aString = 'Input data file was assumed to contain columns q,I,SD'
    print (aString)

if max(data_q_aList) > 1.0:
    aString = 'Data appear to be in reciprocal-nm. Converting to reciprocal-A.'
    print (aString)
    convert_q(data_q_aList)

num_data_true = len(data_i_aList)
qmax_true = max(data_q_aList)
qmin_true = min(data_q_aList)
mean_dq = (qmax_true - qmin_true)/(num_data_true-1)

aString = 'Min q from data input= ' + str("%6.4f"%(qmin_true))
print (aString)
aString = 'Max q from data input= ' + str("%6.4f"%(qmax_true))
print (aString)
aString = 'Number of data points= ' + str(num_data_true)
print (aString)
aString = 'Mean data interval= ' + str("%6.4f"%(mean_dq))
print (aString)

q_for_i_min = test_data_i_min(data_q_aList,data_i_aList,nbin)

##################
# Regularization #
##################

num_guinier = check_guinier(data_q_aList,data_i_aList,qmax_rg,qmin_rg)

if num_guinier > 5:
    aString = 'Number of data in Guinier region= ' + str(num_guinier)
    print (aString)
else:
    aString = '\nStop. Only ' + str(num_guinier) + ' points in the Guinier region'
    print(aString)
    time.sleep(5)
    sys.exit(1)

# Calculate I0,Rg from low-q data using the Guinier equation
(rg_obs,I0_obs) = calc_rg(data_q_aList,data_i_aList,qmax_rg,qmin_rg)

# Get intensity at data edge from last few points
i_limit = statistics.mean(data_i_aList[num_data_true-6:num_data_true-1])
i_ratio = 100.0*i_limit/I0_obs

aString = 'Ratio of I(qmax) to I0= ' + str("%4.3f"%(i_ratio)) + '%'
print (aString)

data_rect_q_aList = []
data_rect_i_aList = []

# Interpolate absent low-q data using Guinier equation
num_rg = 0

q = mean_dq
while q < qmin_true:

    i_predict = get_guinier(rg_obs,I0_obs,q)    
    data_rect_q_aList.append(q)
    data_rect_i_aList.append(i_predict)
    num_rg = num_rg + 1

    q = q + mean_dq

# Replace experimental data on a new regular grid
while q < qmax_true:

    i_interp = get_i_interp(data_q_aList,data_i_aList,q)    
    data_rect_q_aList.append(q)
    data_rect_i_aList.append(i_interp)

    q = q + mean_dq

#################################################
# Generate a filtered P(r) for dmax estimation  #
#################################################

r_aList = []
pr_aList = []
pr_sd_aList = []
data_rect_filter_i_aList = []
r_aList.append(0.0)
pr_aList.append(0.0)
pr_sd_aList.append(0.0)
qmax_filter = min(0.2,0.5*qmax_true)

get_filtered_i(data_rect_filter_i_aList,data_rect_i_aList,data_rect_q_aList,qmax_filter)

num_data_filter = len(data_rect_i_aList)
mean_min_i = statistics.mean(data_rect_filter_i_aList[num_data_filter-6:num_data_filter-1])
shift_background(data_rect_filter_i_aList,mean_min_i)

i = 1
while i < r_max:
    r_aList.append(rstep*float(i))
    i = i + 1

pr_aList.append(0.0)
pr_sd_aList.append(0.0)

i = 1
while i < r_max:
    r = r_aList[i]
    pr = calc_pr(data_rect_q_aList,data_rect_filter_i_aList,mean_dq,r)
    pr_aList.append(pr)
    pr_sd_aList.append(0.0)
  
    i = i + 1

normalize_pr(r_aList,pr_aList,pr_sd_aList,set_pr_scale)
r_limit = len(r_aList)
write_pr(pr_file_filter,r_aList,pr_aList,pr_sd_aList,r_limit)

# Estimate a minimal dmax from filtered P(r)
dmax_est = find_dmax(i_cut,r_aList,pr_aList,rg)

if dmax_est == 0.0:
    aString = 'Dmax could not be determined from filtered P(r)'
    print (aString)

#####################################################################
# Data extrapolation using q-4 function and background subtraction  #
#####################################################################
q_add = 0
num_add = 0

num_data = len(data_rect_i_aList)
mean_min_i = statistics.mean(data_rect_i_aList[num_data-6:num_data-1])

if mean_min_i < 0.0:
    aString = 'Stop. Reduce q range. Intensities at the high-q edge are less than zero.'
    print (aString)
    time.sleep(5)
    sys.exit(1)

(num_add,q_add) = q4_extend(data_rect_i_aList,data_rect_q_aList,\
                     mean_dq,i_ratio_end)

num_data = len(data_rect_i_aList)
min_i_back = data_rect_i_aList[num_data-1]
shift_background(data_rect_i_aList,min_i_back)
    
q_min_reg = min(data_rect_q_aList)
q_max_reg = max(data_rect_q_aList)

aString = 'Input Rg= ' + str("%4.1f"%(rg)) + ' Recalculated Rg= ' + str("%4.1f"%(rg_obs))
print (aString)
aString = 'No of low-q data points interpolated= ' + str(num_rg)
print (aString)
aString = 'No of high-q data points extrapolated= ' + str(num_add)
print (aString)
aString = 'Number of regrid data= ' + str(num_data)
print (aString)
aString = 'Min q for regrid data= ' + str("%6.4f"%(q_min_reg))
print (aString)
aString = 'Max q for extrapolated regrid data= ' + str("%6.4f"%(q_max_reg))
print (aString)
aString = 'Estimated dmax from smoothed P(r)= ' + str("%4.1f"%(dmax_est))
print (aString)

if dmax_in != 0.0:
    aString = 'Using input dmax= ' + str("%4.1f"%(dmax_in))
    print (aString)
    dmax = dmax_in 
else:
    aString = 'Using estimated dmax'
    print (aString)
    dmax = dmax_est

dmax_pt = round(dmax/rstep)

# Apply post-filter for cases where intensity data is ascending

if (qmax_true - q_for_i_min) > dq_i_min:

    data_rect_filter_i_aList = []
    get_filtered_i(data_rect_filter_i_aList,data_rect_i_aList,data_rect_q_aList,q_for_i_min)

    i = 0
    while i < num_data:
        data_rect_i_aList[i] = data_rect_filter_i_aList[i]
        i = i  + 1

    aString = 'Applied Butterworth filter with qthresh= ' + str("%6.4f"%(q_for_i_min))
    print(aString)
else:
    q_for_i_min = 0.0

###################
# Calculate P(r)  #
###################

r_aList = []
pr_aList = []
pr_sd_aList = []
r_aList.append(0.0)
pr_aList.append(0.0)
pr_sd_aList.append(0.0)

i = 1
while i < r_max:
    r_aList.append(rstep*float(i))
    i = i + 1

i = 1
while i < r_max:
    r = r_aList[i]
    pr = calc_pr(data_rect_q_aList,data_rect_i_aList,mean_dq,r)
    pr_aList.append(pr)
    pr_sd_aList.append(0.0)
  
    i = i + 1

##########################################################
# Estimate P(r) error from MC sampling of sd-based error #
##########################################################
k = 0
while k < num_trial:

    i_sd_aList = []
    i = 0
    while i < num_data_true:
        i_sd_aList.append(random.gauss(0.0,data_sd_aList[i]))
        i = i + 1

    i = 1
    while i < r_max:
        r = r_aList[i]
        pr_sd = calc_pr(data_q_aList,i_sd_aList,mean_dq,r)
        pr_sd_aList[i] = pr_sd_aList[i] + pr_sd**2
        i = i + 1

    k = k + 1
    
i = 1
while i < r_max:
    pr_sd_aList[i] = math.sqrt(pr_sd_aList[i]/num_trial)
    i = i + 1
        
##################
# Checks on P(r) #
##################

# Write initial unmodified P(r)
r_limit = len(r_aList)
normalize_pr(r_aList,pr_aList,pr_sd_aList,set_pr_scale)
write_pr(pr_file_start,r_aList,pr_aList,pr_sd_aList,r_limit)

# Modify P(r) by smoothing to zero near dmax and zero beyond dmax
smooth_dmax(pr_aList,dmax_pt)
pr_bound(pr_aList,r_aList,dmax_pt)

# Write modified P(r) limited to bounded region
normalize_pr(r_aList,pr_aList,pr_sd_aList,set_pr_scale)
write_pr(pr_file_mod,r_aList,pr_aList,pr_sd_aList,dmax_pt+1)

rg_pr = get_rg(pr_aList,r_aList,dmax)

# Compute and analyse back transform from final P(r)
data_i_calc_aList = []
ft_to_intensity(data_q_aList,data_i_calc_aList,r_aList,pr_aList)
(a,b) = scale_i(data_i_aList,data_i_calc_aList,data_q_aList,qmax)
apply_scale(data_i_aList,data_i_calc_aList,a,b)
chisq = get_chisq(data_i_aList,data_i_calc_aList,data_sd_aList)
r_factor = get_r_factor(data_i_aList,data_i_calc_aList)

aString = 'Rg from P(r)= ' + str("%4.1f"%(rg_pr))
print (aString)
aString = 'R-factor= ' + str("%6.4f"%(r_factor)) + ' ChiSQ= ' + str("%5.2f"%(chisq))
print(aString)

write_i(i_file,data_q_aList,data_i_aList,data_i_calc_aList)

write_gnomlike(i_and_pr_file,\
                data_q_aList,data_i_aList,data_sd_aList,data_i_calc_aList,\
                r_aList,pr_aList,pr_sd_aList,dmax_pt+1)

#########################################
# (option to) compare P(r) to reference #
#########################################

if aFile_in_pr != 'none':

    r_ref_aList = []
    pr_ref_aList = []
    pr_sd_ref_aList = []

    read_ref_pr(aFile_in_pr,r_ref_aList,pr_ref_aList,pr_sd_ref_aList)

    # Make sure angstroms not nano
    num_ref = len(r_ref_aList)

    if r_ref_aList[1] < 0.2:

        aString = '\nReference P(r) appears to be in nanometers - Converting to A'
        print (aString)
        i = 0
        while i < num_ref:
            r_ref_aList[i] = 10.0*r_ref_aList[i]
            i = i + 1

    ref_step = r_ref_aList[1]
    
    aString = '\nGrid step for reference P(r)= ' + str(ref_step)
    print (aString)

    # Necessary length of new grid
    num_new_pt = int(dmax/ref_step + 1.0)
    
    # If reference is shorter than P(r) pad it to length needed

    num_pad = num_new_pt - num_ref + 1
    if num_pad > 0:

        r = r_ref_aList[num_ref-1] + ref_step
        i = 0
        while i < num_pad + 1:
            r_ref_aList.append(r)
            pr_ref_aList.append(0.0)
            pr_sd_ref_aList.append(0.0)
            r = r + ref_step

            i = i + 1
        
    num_ref = len(r_ref_aList)

    # Recompute P(r) onto the grid of the reference
    pr_aList = []
    pr_sd_aList = []
    pr_aList.append(0.0)
    pr_sd_aList.append(0.0)

    i = 1
    while i < num_ref:
        r = r_ref_aList[i]
        pr = calc_pr(data_rect_q_aList,data_rect_i_aList,mean_dq,r)
        pr_aList.append(pr)
        pr_sd_aList.append(0.0)
        i = i + 1

    # Smooth and bound
    smooth_dmax(pr_aList,num_new_pt)
    pr_bound(pr_aList,r_ref_aList,num_new_pt)
    
    normalize_pr(r_ref_aList,pr_aList,pr_sd_aList,set_pr_scale)
    normalize_pr(r_ref_aList,pr_ref_aList,pr_sd_ref_aList,set_pr_scale)

    # Compare
    dif_pr = compare_pr(pr_aList,pr_ref_aList)

    aString = 'Difference from reference P(r)= ' + str("%4.2f"%(100.0*dif_pr)) + '%'
    print (aString)
    
    write_pr(pr_file_with_ref,r_ref_aList,pr_aList,pr_ref_aList,num_ref)

####################
# Report and close #
####################

write_statistics(log_file,aString_time,sample_id,qmin_true,qmax_true,q_for_i_min,i_ratio,dmax_est,\
                 dmax,r_factor,chisq,rg_pr,aFile_in,aFile_in_pr,dif_pr)

if dirname != '':
    aString = '\nOutput files are in folder= ' + str(dirname) + '\n'
    print (aString)

aString = '\nProgram run completed'
print (aString)
aString  = 'Closing...'
print (aString)
time.sleep(8)
sys.exit(1)
